/* 
   d3.cpp

   3d-classes (vectors, lines, planes, rotation)
 
     Copyright (C) 2005, Patrick Rammelt <rammelt@cs.tu-berlin.de>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
*/

#ifndef _D3_CPP
#define _D3_CPP


#include <math.h>
#include "mathematical.cpp"

#define EPS 0.00000001

#ifdef _NOISINF
#define isinf(x) (0)
#endif


/* =========================================================================
   3d-object
   ========================================================================= */
class object3d 
{
public:
  virtual ~object3d (void) { }
  virtual bool isvalid (void) const { return(1); }
};


/* =========================================================================
   3d-vector 
   for operations in 3d euclidean space 
   ========================================================================= */
class vector3d : public object3d
{
public:

  double x;
  double y; 
  double z;

public:

  // construct / destruct
  vector3d (void) { }
  vector3d (double v) { set (v); }
  vector3d (const class vector3d &v) { set (v); }
  vector3d (const class vector3d &v, double l) { set (v,l); }
  vector3d (double x0, double y0, double z0) { set (x0,y0,z0); }
  virtual ~vector3d (void) { }

  // set
  virtual inline vector3d& set (double v) { 
    x = y = z = v; 
    return (*this);
  }
  virtual inline vector3d& set (const class vector3d &v) { 
    return (set (v.x, v.y, v.z)); 
  }
  virtual inline vector3d& set (const class vector3d &v, double l) { 
    set (v); 
    norm (l); 
    return (*this);
  }
  virtual inline vector3d& set (double x0, double y0, double z0) { 
    x = x0; 
    y = y0; 
    z = z0; 
    return (*this);
  }

  // set to reverse v
  virtual vector3d& reverse (void) { 
    return (set (-x, -y, -z));
  }
  virtual vector3d& reverse (const class vector3d &v) { 
    return (set (-v.x, -v.y, -v.z)); 
  }

  // make angle more / less than 90 degrees
  virtual vector3d& along (const class vector3d &v) {
    class vector3d s (x+v.x, y+v.y, z+v.z); 
    if (s.length()>(length()+v.length())) reverse(); 
    return (*this);
  }
  virtual vector3d& oppose (const class vector3d &v) {
    class vector3d s (x+v.x,y+v.y,z+v.z); 
    if (s.length()<(length()+v.length())) reverse(); 
    return (*this);
  }

  // set distance to ref to d (stay on line this-ref)
  virtual vector3d& setdist (const class vector3d &ref, double d) { 
    if (*this!=ref) { 
      *this -= ref; 
      *this *= d/length(); 
      *this += ref; 
    } 
    return (*this);
  }

  // valid check
  virtual bool isvalid (void) const { 
    return (!(isnan(x) || isnan(y) || isnan(z) || 
	      isinf(x) || isinf(y) || isinf(z))); 
  }

  // euclidean length
  virtual double sqlen  (void) const { 
    return ((x*x)+(y*y)+(z*z)); 
  }
  virtual double length (void) const { 
    return (sqrt(sqlen())); 
  }

  // euclidean dist |this-v|
  virtual double sqdist (const class vector3d &v) const {
    double dx=x-v.x, dy=y-v.y, dz=z-v.z; 
    return ((dx*dx)+(dy*dy)+(dz*dz)); 
  }
  virtual double dist (const class vector3d &v) const {
    return (sqrt(sqdist(v))); 
  }

  // norm: length=1 / length len
  virtual vector3d& norm (void) { 
    return (*this /= length()); 
  }
  virtual vector3d& norm (double len) { 
    return (*this *= (len/length())); 
  }

  // spin around Z-axis by angle alpha
  virtual vector3d& spinZ (double alpha) {
    double x0 = x; 
    x = x0*cos(alpha) - y*sin(alpha); 
    y = x0*sin(alpha) + y*cos(alpha); 
    return (*this);
  }

  // spin around Y-axis by angle alpha
  virtual vector3d& spinY (double alpha) {
    double z0 = z; 
    z = z0*cos(alpha) - x*sin(alpha); 
    x = z0*sin(alpha) + x*cos(alpha); 
    return (*this);
  }

  // spin around X-axis by angle alpha
  virtual vector3d& spinX (double alpha) {
    double y0 = y; 
    y = y0*cos(alpha)-z*sin(alpha); 
    z = y0*sin(alpha)+z*cos(alpha); 
    return (*this);
  }

  // angle between vector u and v (lines g1:x=a+t*u and g2:x=b+t*v)
  virtual double angle (const class vector3d &v) const { 
    return (acos(cosinus(v))); 
  }

  // cosinus of angle between vector u and v (lines g1:x=a+t*u and g2:x=b+t*v)
  virtual double cosinus (const class vector3d &v) const {
    return ((*this*v) / (length()*v.length())); 
  }

  // point of intersection : line g:x=a+t*b - plane E:n*x=d
  virtual void intersection (const class vector3d &a, const class vector3d &b, 
			     const class vector3d &n, double d) {
    double t0a = d - (n*a);
    double t0b = n*b;            // t0b==0 => no intersection => !valid()
    set (a + (b*(t0a/t0b))); 
  }

  // nearest point on line g:a+t*b to point p
  virtual void projection (const class vector3d &p,
			   const class vector3d &a, const class vector3d &b) {
    class vector3d nb (b,1.0);
    double d = (nb*p);            // plane E:b*x=d
    intersection (a,nb,nb,d);
  }

  // set
  class vector3d& operator = (const class vector3d &v) {
    return (set (v));
  }
  class vector3d& operator = (double v) { 
    return (set (v)); 
  }

  // equality / inequality
  bool operator == (const class vector3d &v) const {
    return (x==v.x && y==v.y && z==v.z); 
  }
  bool operator != (const class vector3d &v) const { 
    return (!(*this==v)); 
  }

  // dot operations
  class vector3d& operator += (const class vector3d &v) { 
    return (set (x+v.x, y+v.y, z+v.z));
  }
  class vector3d& operator -= (const class vector3d &v) { 
    return (set (x-v.x, y-v.y, z-v.z));
  }
  class vector3d& operator *= (const class vector3d &v) { 
    return (set (x*v.x, y*v.y, z*v.z));
  }
  class vector3d& operator /= (const class vector3d &v) { 
    return (set (x/v.x, y/v.y, z/v.z));
  }

  // cross-product vector x vector
  class vector3d operator ^= (const class vector3d &v) { 
    return (set (y*v.z - z*v.y,
		 z*v.x - x*v.z,
		 x*v.y - y*v.x));
  }

  // operations vector scalar
  class vector3d& operator += (double d) { 
    return (set (x+d, y+d, z+d)); 
  }
  class vector3d& operator -= (double d) { 
    return (set (x-d, y-d, z-d)); 
  }
  class vector3d& operator *= (double d) { 
    return (set (x*d, y*d, z*d)); 
  }
  class vector3d& operator /= (double d) { 
    return (set (x/d, y/d, z/d)); 
  }

  // operations vector - scalar
  class vector3d operator + (double d) const { 
    vector3d w(x+d, y+d, z+d);
    return (w); 
  }
  class vector3d operator - (double d) const { 
    vector3d w(x-d, y-d, z-d);
    return (w); 
  }
  class vector3d operator * (double d) const { 
    vector3d w(x*d, y*d, z*d);
    return (w); 
  }
  class vector3d operator / (double d) const { 
    vector3d w(x/d, y/d, z/d);
    return (w); 
  }
  
  // operations vector - vector
  class vector3d operator + (const class vector3d &v) const { 
    vector3d w(x+v.x, y+v.y, z+v.z);
    return (w); 
  }
  class vector3d operator - (const class vector3d &v) const { 
    vector3d w(x-v.x, y-v.y, z-v.z);
    return (w); 
  }
  
  // scalar-product vector x vector
  double operator * (const class vector3d &v) const { 
    return (x*v.x + y*v.y + z*v.z); 
  }
  
  // cross-product vector x vector 
  // (creates a vector which is orthogonal to both input vectors)
  class vector3d operator ^ (const class vector3d &v) const { 
    vector3d w(y*v.z - z*v.y,
	       z*v.x - x*v.z,
	       x*v.y - y*v.x); 
    return (w);
  }
};


/* =========================================================================
   3d-line g:x=a+t*b
   ========================================================================= */
class line3d : public object3d
{
public:

  class vector3d a;
  class vector3d b;

public:

  // construct / destruct
  line3d (void) {}
  line3d (const class vector3d &p, const class vector3d &n) { 
    a = p; b.set (n,1.0); 
  }
  line3d (const class line3d &l) { 
    a = l.a; 
    b = l.b; 
  }
  virtual ~line3d (void) { }

  // set
  virtual inline line3d& set (const class vector3d &p, const class vector3d &n) { 
    a = p; 
    b.set (n,1.0); 
    return (*this);
  }
  virtual inline line3d&  set (const class line3d &l) { 
    a = l.a; 
    b = l.b; 
    return (*this);
  }

  // valid check
  virtual bool isvalid (void) const { 
    return (a.isvalid() && b.isvalid()); 
  }

  // square dist of point p to line
  virtual double sqdist (const class vector3d &p) const {
    class vector3d pa;  pa  = p-a;
    class vector3d pab; pab = p-(a+b);
    double x=b.length(), y=pa.length(), z=pab.length();
    return ((x*x*x*x + y*y*y*y + z*z*z*z - 
	     2.0*x*x*y*y - 2.0*x*x*z*z - 2.0*y*y*z*z) / (-4.0*x*x));
  }

  //  0: point p is on plane E given by point a and normal vector b
  //  1: p is in direction  b to plane E
  // -1: p is in direction -b to plane E
  virtual int direction (const class vector3d &p) const {
    double d = (b*a) - (b*p);
    if (d>0.0) return (1); else if (d<0.0) return (-1); else return (0); }

  // distance of point p to line
  virtual double dist (const class vector3d &p) const { 
    return (sqrt(MAX(sqdist(p),0.0))); }

  // make this a intersection of two planes 
  // line direction is just n1^n2, to get one point set one non-zero 
  // (the biggest) b-element to 0 and solve for the other two
  virtual line3d& intersection (const class vector3d &n1, double d1,
			     const class vector3d &n2, double d2) {
    b.set (n1^n2, 1.0);
    if (ABS(b.x)>ABS(b.y) && ABS(b.x)>ABS(b.z)) { 
      a.x=0; solve (n1.y, n1.z, d1,  n2.y, n2.z, d2,  a.y, a.z); 
    } else if (ABS(b.y)>ABS(b.z)) { 
      a.y=0; solve (n1.x, n1.z, d1,  n2.x, n2.z, d2,  a.x, a.z); 
    } else {
      a.z=0; solve (n1.x, n1.y, d1,  n2.x, n2.y, d2,  a.x, a.y); 
    }
    return (*this);
  }

  // nearest point on line lp to point p
  virtual class vector3d projection (const class vector3d &p) const {
    class vector3d lp; 
    projection(p,lp); 
    return (lp); 
  }
  virtual void projection (const class vector3d &p, class vector3d &lp) const {
    lp.projection(p,a,b); 
  }

  // set
  class line3d operator = (const class line3d &l) { 
    return (set (l));
  }

protected:

  // a1*x + b1*y = c1
  // a2*x + b2*y = c2
  void solve (double a1, double b1, double c1, 
	      double a2, double b2, double c2,
	      double &x, double &y) const {
    y = ((a2*c1)-(a1*c2)) / ((a2*b1)-(a1*b2));
    x = (ABS(a1)>ABS(a2) ? (c1-(b1*y)) / a1 : (c2-(b2*y)) / a2);
  }
};


/* =========================================================================
   3d-plane E:n*x=d
   ========================================================================= */
class plane3d : public object3d
{
public:

  class vector3d n;
  double d;

public:

  // construct / destruct
  plane3d (void) {}
  plane3d (const class vector3d &norm, double dist) { 
    set (norm,dist); 
  }
  plane3d (const class plane3d &p) { 
    n = p.n; 
    d = p.d; 
  }
  plane3d (const class vector3d &norm, const class vector3d &p) { 
    set (norm,p); 
  }
  plane3d (const class vector3d &a, 
	   const class vector3d &u,
	   const class vector3d &v) { 
    set (a,u,v); 
  }

  // set
  virtual inline plane3d& set (const class vector3d &norm, double dist) { 
    n.set (norm,1.0); 
    d = dist; 
    return (*this);
  }
  virtual inline plane3d& set (const class plane3d &p) { 
    n = p.n; 
    d = p.d; 
    return (*this);
  }

  // valid check
  virtual bool isvalid (void) const { 
    return (n.isvalid() && !(isnan(d) || isinf(d))); 
  }

  // get normal-form (E:n*x=d) from normal-vector norm and one point p
  virtual void set (const class vector3d &norm, const class vector3d &p) {
    n.set(norm,1.0); d = n*p; 
  }

  // get normal-form (E:n*x=d) from 1 point and 2 direction vectors
  virtual void set (const class vector3d &a, 
		    const class vector3d &u, 
		    const class vector3d &v) { 
    n.set(u^v,1.0); 
    d = (n*a); 
  }

  // angle between vector p and plane 
  virtual double angle (const class vector3d &p) const { 
    return ( asin((n*p) / (p.length()*n.length())) ); 
  }

  // point of intersection : line g:x=a+t*b
  virtual class vector3d intersection (const class line3d l) const {
    class vector3d i; 
    intersection (l,i); 
    return (i); 
  }
  virtual void intersection (const class line3d l, class vector3d &i) const {
    i.intersection (l.a,l.b,n,d); 
  }

  // line of intersection : plane E2:n2*x=d2
  virtual class line3d intersection (const class plane3d &e) const {
    class line3d l; 
    intersection (e,l); 
    return (l); 
  }
  virtual void intersection (const class plane3d &e, class line3d &l) const {
    l.intersection (n,d,e.n,e.d); 
  }

  // nearest point on plane to point in 3d space
  virtual class vector3d projection (const class vector3d &v) const {
    class vector3d p; 
    projection (v,p); 
    return (p); 
  }
  virtual void projection (const class vector3d &v, class vector3d &p) const {
    class line3d l (v,n); 
    intersection (l,p); 
  }
  
  // mirroring a point
  virtual class vector3d mirror (const class vector3d &v) const {
    class vector3d m; 
    mirror (v,m); 
    return (m); 
  }
  virtual void mirror (const class vector3d &v, class vector3d &m) const {
    m.set (v - (n*(2.0*dist(v)/n.length()))); 
  }

  // reflecting a line
  virtual class line3d reflection (const class line3d &l) const {
    class line3d r; 
    reflection (l,r); 
    return (r); 
  }
  virtual void reflection (const class line3d &l, class line3d &r) const {
    class vector3d p; 
    projection   (l.a,p);
    intersection (l,r.a);
    r.b = r.a-(p+p)+l.a;
  } 

  // dist point p to plane E:n*x=d
  virtual double dist (const class vector3d &p) const { 
    return ((n*p) - d); 
  }

  // set
  class plane3d& operator = (const class plane3d &e) { 
    return (set (e));
  }
};


/* =========================================================================
   rotation of vectors in 3d euclidean space 
   ========================================================================= */
class rotation3d {

protected:

  double cosX; 
  double sinX;
  double cosY;
  double sinY;

public:

  rotation3d (void) : cosX(1), sinX(0), cosY(1), sinY(0) {}
  rotation3d (const class rotation3d &r) { set(r); }
  rotation3d (const class vector3d &v)   { set(v); }
  virtual ~rotation3d (void) {}

  virtual inline rotation3d& set (const class rotation3d &r) { 
    cosX = r.cosX; 
    sinX = r.sinX; 
    cosY = r.cosY; 
    sinY = r.sinY; 
    return (*this);
  }

  // set rotation direction to v
  virtual inline rotation3d& set (const class vector3d &w) { 
    class vector3d v(w);
    
    v.z = -v.z;                  // z is getting smaller when getting closer 
    v.x = -v.x;
    v.y = -v.y;

    double lxz = sqrt(v.z*v.z + v.x*v.x);
    if (lxz==0) { 
      cosY=1; 
      sinY=0; 
    } else {
      cosY =   v.z/lxz;          // cos(-alpha) = cos(alpha)
      sinY = -(v.x/lxz);         // sin(-alpha) = -sin(alpha)
    }

    double z0=v.z; 
    v.z=z0*cosY-v.x*sinY; v.x=z0*sinY+v.x*cosY;  // rotate

    double lyz = sqrt(v.y*v.y + v.z*v.z);
    if (lyz==0) { 
      cosX=1; 
      sinX=0; 
    } else {
      cosX =   v.z/lyz;          // cos(-alpha) = cos(alpha)
      sinX = -(v.y/lyz);         // sin(-alpha) = -sin(alpha)
    }

    return (*this);
  }

  // rotate vector p by pre-set angles (for rotation of view point)
  void rotate (vector3d &p) const {
    double z0 = p.z; 
    p.z = z0*cosX-p.y*sinX; 
    p.y = z0*sinX+p.y*cosX; 
    z0 = p.z; 
    p.z = z0*cosY-p.x*sinY; 
    p.x = z0*sinY+p.x*cosY; 
  }

};


#endif // ifndef _D3_CPP

